import numpy as np
import _pickle as pickle
import gym

from chainer import cuda
import time, threading
import sys
import math
import copy
import scipy

# for visualization
from scipy.misc import toimage
from scipy.stats import entropy

# import our dqn agent
from simple_dqn_agent import SimpleDQN

# some common util functions
from img_preprocessing import *
#from buffer_queue import QueueBuffer

# some constants
# DOMAIN = 'Atlantis-v4'
DOMAIN = 'KungFuMaster-v4'
REWARD_SURVIVE = 1.0 # how much reward to give just for surviving a step

# input image size
IMAGE_WIDTH = 84
IMAGE_HEIGHT = 84
GRAYSCALE = False
DOWNSAMPLE_FACTOR = 1 # keep this as one

# network params
NUM_HIDDEN = 200
GAMMA = 0.99
LEARNING_RATE = 1e-3
DECAY_RATE = 0.99

# exploration params
MAX_EPSILON = 0.2
MIN_EPSILON = 0.05
EXPLORATION_STOP = 500000   # at this step epsilon will be GREEDY_EPS_MIN
LAMBDA = - math.log(0.01) / EXPLORATION_STOP  # speed of decay

# MDL constants
MDL_STATE_BUFFER_SIZE = 40
MDL_MIN_BACK = 20 # do not go back less than 20 steps back

def processImage( img ):
	
	# preprocess based on domain
	img = prepro(img,DOMAIN,DOWNSAMPLE_FACTOR,GRAYSCALE)
	
	# resize based on constants
	img = scipy.misc.imresize(img, (IMAGE_WIDTH, IMAGE_HEIGHT), interp='bilinear')
	
	#toimage(img).save("test_crop_gray.png")
	
	return img
   
# Environment class
class Environment:
	def __init__(self, problem):
		self.problem = problem
		self.env = gym.make(problem)
		self.total_game_steps = 0		

	# plays an episode
	def run(self, dqn_agent, evaluate=False):                
		# whether or not to explore
		explore = True if evaluate==False else False
	
		# get initial observation
		observation = self.env.reset()
			
		# process image
		cur_I = processImage(observation)
		
		# compute input state
		prev_img = None
		cur_img = cur_I.astype(np.float).ravel()

		# clear set of subtasks
		self.subtasks = []

		R = 0
		r = 0
		while True:  
			
			# set exploration rate
			if explore:
				epsilon = MIN_EPSILON + (MAX_EPSILON - MIN_EPSILON) * math.exp(-LAMBDA * self.total_game_steps)
				#print(epsilon)
				dqn_agent.set_explore_epsilon(epsilon)
			    
			# render
			self.env.render()

			# store prev_img
			x = cur_img - prev_img if prev_img is not None else np.zeros(D)
			prev_img = cur_img
			
			# act and perceive
			a = dqn_agent.process_step(x,explore)
			observation, r, done, info = self.env.step(a)
			cur_img = processImage(observation).astype(np.float).ravel()
		
			# store reward
			R += r
			
			# clip reward and give to agent -- ideally, the reward should be
			# scaled between -1 and 1 where the smallest possible reward to be
			# observed in the game is -1 and the largest possible reward is 1
			# r = np.clip(r, -1, 1)   # clip reward to [-1, 1]
			
			if r == 0:
				r = REWARD_SURVIVE # give it reward just for surviving
							
			agent.give_reward(r)
  
			if done: # terminal state
				if evaluate==False:
					#print("Finishing episode...")
					agent.finish_episode() # also resets
					
				else:
					#print("Reseting agent...")
					agent.reset()
				x = None
				cur_img = None
				pre_img = None
	       
			if evaluate == False:
				self.total_game_steps += 1 

			if done:
				return [self.total_game_steps,R]
				break
				
#-------------------- MAIN ----------------------------

# get aruments
test_num = int(sys.argv[1])  # used for the random seed and saving file names
max_num_game_steps = int(sys.argv[2])
exp_tag = sys.argv[3]

# set random seed
random_seed = test_num

# save filename
curve_file_name = exp_tag+"_"+str(random_seed)+".txt"
  
# create our environment
env = Environment(DOMAIN)

# get action count
actionCnt = env.env.action_space.n

# create our agent
D = IMAGE_HEIGHT * IMAGE_WIDTH * 3 if GRAYSCALE == False else IMAGE_HEIGHT * IMAGE_WIDTH
agent = SimpleDQN(actionCnt,D,NUM_HIDDEN,LEARNING_RATE,GAMMA,DECAY_RATE,MAX_EPSILON,random_seed)

# training params -- how many episodes to play before updating parameters
eps_per_epoch = 5

# counters
current_epoch = 0
steps_taken = 0

running_reward = None
ep_reward_list = []

while True:
	# evaluate
	eval_result = env.run(agent,evaluate=True)
	total_steps = eval_result[0]
	eval_reward = eval_result[1]

	running_reward = eval_reward if running_reward is None else running_reward * 0.9 + eval_reward * 0.1
	print(str(total_steps)+","+str(eval_reward)+","+str(running_reward))
	
	ep_reward_list.append([current_epoch,total_steps,eval_reward,running_reward])
	save_curve(curve_file_name,ep_reward_list)
	
	for row in ep_reward_list:
		print("%f,%f,%f,%f" % (row[0],row[1],row[2],row[3]))

	
	if env.total_game_steps > max_num_game_steps:
		break
	
	# play a few episodes
	for i in range(0,eps_per_epoch):
		#print("Train episode #" + str(i+eps_per_epoch*current_epoch))
		
		# play regular episode
		train_result = env.run(agent)
		
	
	current_epoch += 1
		
	# update parameters
	#print("Updating parameters...")
	agent.update_parameters()
	
		



	




